/*
 * Copyright (c) 2009-2012 LodgON  (http://www.lodgon.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
package com.lodgon.dali.module.forum.ejb;

import com.lodgon.dali.core.ejb.AuthorizationBean;
import com.lodgon.dali.core.ejb.ContentBean;
import com.lodgon.dali.core.ejb.DaliCoreException;
import com.lodgon.dali.core.ejb.GroupBean;
import com.lodgon.dali.core.entity.Content;
import com.lodgon.dali.core.entity.ContentPermission;
import com.lodgon.dali.core.entity.Group;
import com.lodgon.dali.core.entity.TypePermission;
import com.lodgon.dali.core.entity.User;
import com.lodgon.dali.core.entity.search.ContentSearchParams;
import com.lodgon.dali.module.forum.Constants;
import com.lodgon.dali.module.forum.ForumException;
import com.lodgon.dali.module.forum.entity.*;
import java.util.LinkedList;
import java.util.List;
import javax.ejb.EJB;
import javax.ejb.Stateless;
import javax.persistence.EntityManager;
import javax.persistence.PersistenceContext;
import javax.persistence.Query;

@Stateless
public class ForumBean {

  @PersistenceContext
  EntityManager em;
  @EJB
  AuthorizationBean authorizationBean;
  @EJB
  ContentBean contentBean;
  @EJB
  GroupBean groupBean;

  /**
   * Create a Forum. The author of the Forum should have TypePermission with
   * action FORUM_CREATE
   *
   * @param forum the forum that has to be created, including a valid author
   * field
   * @return the managed Forum object
   * @throws DaliCoreException in case the author has no permission to create a
   * forum
   */
  public Forum createForum(Forum forum) throws DaliCoreException {
    User author = forum.getAuthor();
    if (hasTypePermission(Constants.TYPE_PERMISSION_FORUM, author, Constants.FORUM_PERMISSION_ACTION_CREATE_FORUM, forum.getAppKey())) {
      return (Forum) contentBean.create(forum);
    } else {
      throw new DaliCoreException(DaliCoreException.NO_PERMISSION, "No permission to create forum.");
    }
  }

  public Forum getForum(String appKey, User user, String forumUid) throws DaliCoreException {
    Forum forum = (Forum) contentBean.getByUid(appKey, forumUid);
    if (forum == null) {
      return null;
    }
    if (hasPermission(forum.getId(), user, Constants.FORUM_PERMISSION_ACTION_VIEW, appKey)) {
      System.out.println("HAVE PERMISSION");
      return forum;
    } else {
      System.out.println("HAVE NO PERMISSION");
      throw new DaliCoreException(DaliCoreException.NO_PERMISSION, "No permission to view forum.");
    }
  }

  /**
   * Retrieve all Forums for the given application.
   *
   * @param appKey
   * @param user
   * @return
   * @throws DaliCoreException
   */
  public List<Forum> getForums(String appKey, User user) throws DaliCoreException {
    Query q = em.createNamedQuery("Forum.listForums");
		q.setParameter("appKey", appKey);
    List<Forum> answer = q.getResultList();
    return answer;
  }

  /**
   * Create a Category. It is required that the
   * <code>getForum</code> method on the supplied Category returns a valid
   * <code>Forum</code> instance, otherwise a
   * <code>ForumException.NO_PARENT</code> exception will be thrown. <br/> The
   * author of the supplied
   * <code>category</code> should have
   * <code>CREATE_CATEGORY</code> permission on the supplied forum.
   *
   * @param category the category that has to be made persistent, including a
   * parent forum and a author user.
   * @return the managed entity object.
   * @throws DaliCoreException
   */
  public Category createCategory(Category category) throws DaliCoreException {
    Forum parent = category.getForum();
    if (parent == null) {
      throw new ForumException(ForumException.NO_PARENT);
    }
    User author = category.getAuthor();
    if (hasPermission(parent.getId(), author, Constants.FORUM_PERMISSION_ACTION_CREATE_CATEGORY, category.getAppKey())) {
      category.setParent(parent);
      return (Category) contentBean.create(category);
    } else {
      throw new DaliCoreException(DaliCoreException.NO_PERMISSION, "No permission to create category");
    }
  }

  /**
   * Retrieve the category with the supplied uuid. The user requesting this
   * category should have
   * <code>ACTION_VIEW</code> permission.
   *
   * @param appKey
   * @param user the user requesting this category
   * @param categoryUid
   * @return the category with the supplied uuid, or <code>null</code> in case
   * no such category exist.
   * @throws DaliCoreException
   */
  public Category getCategory(String appKey, User user, String categoryUid)
          throws DaliCoreException {
    Category answer = (Category) contentBean.getByUid(appKey, categoryUid);
    if (answer == null) {
      return null;
    }
    if (!hasPermission(answer.getId(), user, Constants.FORUM_PERMISSION_ACTION_VIEW, appKey)) {
      throw new DaliCoreException(DaliCoreException.NO_PERMISSION, "No permission to view category");
    }
    answer = fixCategory(answer);
    return answer;
  }

  /**
   * Retrieve all Categories that have the given parentId as parent.
   *
   * @param appKey
   * @param user
   * @param parentUid
   * @return
   * @throws DaliCoreException
   */
  public List<Category> getCategoriesByParent(String appKey, User user, String parentUid)
          throws DaliCoreException {
    List<Content> search = getContentByParent(appKey, parentUid);
    LinkedList<Category> answer = new LinkedList();
    for (Content candidate : search) {
      if (candidate instanceof Category) {
        answer.add(fixCategory(candidate));
      }
    }
    return answer;
  }

  public Subforum createSubforum(Subforum subforum) throws DaliCoreException {
    Content parent = subforum.getCategory();
    if (parent == null) {
      parent = subforum.getForum();
    }
    if (parent == null) {
      throw new ForumException(ForumException.NO_PARENT,"No parent supplied for subforum");
    }
    User author = subforum.getAuthor();
    if (hasPermission(parent.getId(), author, Constants.FORUM_PERMISSION_ACTION_CREATE_SUBFORUM, subforum.getAppKey())) {
      subforum.setParent(parent);
      return (Subforum) contentBean.create(subforum);
    } else {
      throw new DaliCoreException(DaliCoreException.NO_PERMISSION, "No permission to create subforum");
    }
  }

  public Subforum getSubforum(String appKey, User user, String subforumUid)
          throws DaliCoreException {
    Subforum answer = (Subforum) contentBean.getByUid(appKey, subforumUid);
    if (answer == null) {
      return null;
    }
    if (!hasPermission(answer.getId(), user, Constants.FORUM_PERMISSION_ACTION_VIEW, appKey)) {
      throw new DaliCoreException(DaliCoreException.NO_PERMISSION, "No permission to view subforum");
    }
    return fixSubforum(answer);
  }

  public List<Subforum> getSubforumsByParent(String appKey, User user, String parentUid)
          throws DaliCoreException {
    List<Content> search = getContentByParent(appKey, parentUid);
    LinkedList<Subforum> answer = new LinkedList();
    for (Content candidate : search) {
      if (candidate instanceof Subforum) {
        answer.add(fixSubforum(candidate));
      }
    }
    return answer;
  }

  public Topic createTopic(Topic topic) throws DaliCoreException {
    Content parent = topic.getSubforum();
    if (parent == null) {
      parent = topic.getCategory();
    }
    if (parent == null) {
      parent = topic.getForum();
    }
    if (parent == null) {
      throw new ForumException(ForumException.NO_PARENT);
    }
    User author = topic.getAuthor();
    if (hasPermission(parent.getId(), author, Constants.FORUM_PERMISSION_ACTION_CREATE_TOPIC, topic.getAppKey())) {
      topic.setParent(parent);
      return (Topic) contentBean.create(topic);
    } else {
      throw new DaliCoreException(DaliCoreException.NO_PERMISSION, "No permission to create topic");
    }
  }

  public Topic getTopic(String appKey, User user, String topicUid)
          throws DaliCoreException {
    Topic answer = (Topic) contentBean.getByUid(appKey, topicUid);
    if (answer == null) {
      return null;
    }
    if (!hasPermission(answer.getId(), user, Constants.FORUM_PERMISSION_ACTION_VIEW, appKey)) {
      throw new DaliCoreException(DaliCoreException.NO_PERMISSION, "No permission to view topic");
    }
    return fixTopic(answer);
  }

  public List<Topic> getTopicsByParent(String appKey, User user, String parentUid)
          throws DaliCoreException {
    List<Content> search = getContentByParent(appKey, parentUid);
    LinkedList<Topic> answer = new LinkedList();
    for (Content candidate : search) {
      if (candidate instanceof Topic) {
        answer.add(fixTopic(candidate));
      }
    }
    return answer;
  }

  public Post createPost(Post post) throws DaliCoreException {
    Content parent = post.getTopic();
    if (parent == null) {
      parent = post.getSubforum();
    }
    if (parent == null) {
      parent = post.getCategory();
    }
    if (parent == null) {
      parent = post.getForum();
    }
    if (parent == null) {
      throw new ForumException(ForumException.NO_PARENT);
    }
    User author = post.getAuthor();
    if (hasPermission(parent.getId(), author, Constants.FORUM_PERMISSION_ACTION_CREATE_POST, post.getAppKey())) {
      post.setParent(parent);
      return (Post) contentBean.create(post);
    } else {
      throw new DaliCoreException(DaliCoreException.NO_PERMISSION, "No permission to create post");
    }
  }

  public Post getPost(String appKey, User user, String postUid)
          throws DaliCoreException {
    Post answer = (Post) contentBean.getByUid(appKey, postUid);
    if (answer == null) {
      return null;
    }
    if (!hasPermission(answer.getId(), user, Constants.FORUM_PERMISSION_ACTION_VIEW, appKey)) {
      throw new DaliCoreException(DaliCoreException.NO_PERMISSION, "No permission to view post");
    }

    return fixPost(answer);
  }

  public List<Post> getPostsByParent(String appKey, User user, String parentUid)
          throws DaliCoreException {
    List<Content> search = getContentByParent(appKey, parentUid);
    LinkedList<Post> answer = new LinkedList();
    for (Content candidate : search) {
      if (candidate instanceof Post) {
        answer.add(fixPost(candidate));
      }
    }
    return answer;
  }

  private boolean hasTypePermission(int type, User user, int action, String appKey) throws DaliCoreException {
    boolean hasPermission = false;
    if (user != null) {
      hasPermission = authorizationBean.checkTypePermission(user, type, action);
    }

    if (!hasPermission) {
      Group everyoneGroup = groupBean.getByUid(appKey, "everyone");
      if (everyoneGroup != null) {
        List<TypePermission> typePermissions = authorizationBean.getTypePermissionsByGroup(everyoneGroup.getId(), type);
        for (TypePermission typePermission : typePermissions) {
          if (typePermission.getAction() == action) {
            hasPermission = true;
            break;
          }
        }
      }
    }

    if (!hasPermission && user != null) {
      Group membersGroup = groupBean.getByUid(appKey, "members");
      if (membersGroup != null) {
        List<TypePermission> typePermissions = authorizationBean.getTypePermissionsByGroup(membersGroup.getId(), type);
        for (TypePermission typePermission : typePermissions) {
          if (typePermission.getAction() == action) {
            hasPermission = true;
            break;
          }
        }
      }
    }
    System.out.println("permission asked for " + user + " will return " + hasPermission);
    return hasPermission;

  }

  private boolean hasPermission(int contentId, User user, int action, String appKey) throws DaliCoreException {
    boolean hasPermission = false;
    if (user != null) {
      hasPermission = authorizationBean.checkContentPermission(user.getId(), contentId, action);
    }

    if (!hasPermission) {
      Group everyoneGroup = groupBean.getByUid(appKey, "everyone");
      if (everyoneGroup != null) {
        List<ContentPermission> contentPermissions = authorizationBean.getContentPermissionsByGroup(everyoneGroup.getId(), contentId);
        for (ContentPermission contentPermission : contentPermissions) {
          if (contentPermission.getAction() == action) {
            hasPermission = true;
            break;
          }
        }
      }
    }

    if (!hasPermission && user != null) {
      Group membersGroup = groupBean.getByUid(appKey, "members");
      if (membersGroup != null) {
        List<ContentPermission> contentPermissions = authorizationBean.getContentPermissionsByGroup(membersGroup.getId(), contentId);
        for (ContentPermission contentPermission : contentPermissions) {
          if (contentPermission.getAction() == action) {
            hasPermission = true;
            break;
          }
        }
      }
    }
    if (!hasPermission && (getRecursivePermission(action)>-1)) {
      // in case we need Xpermission, see if we have recursiveXpermission up on the tree
   //   if (action == Constants.FORUM_PERMISSION_ACTION_VIEW) {
        if (hasPermission(contentId, user, getRecursivePermission(action), appKey)) {
          hasPermission = true;
        }
        Content me = contentBean.getById(contentId);
        if (!hasPermission && ((me != null) && (me.getParentId() != null))){
          Content parent = contentBean.getById(me.getParentId());
          while (parent !=null) {
            if (hasPermission(parent.getId(), user,getRecursivePermission(action), appKey)) {
              hasPermission = true;
              break;
            }
            Integer pid = parent.getParentId();
            if (pid != null) {
              parent = contentBean.getById(parent.getParentId());
            }
            else {
              parent = null;
            }
          }
        }
     // }
    }
    System.out.println("permission asked for " + user + " will return " + hasPermission);
    return hasPermission;
  }

  private int getRecursivePermission(int base) {
    switch (base) {
      case Constants.FORUM_PERMISSION_ACTION_VIEW:
        return Constants.FORUM_PERMISSION_ACTION_VIEW_RECURSIVE;
      case Constants.FORUM_PERMISSION_ACTION_CREATE_CATEGORY:
        return Constants.FORUM_PERMISSION_ACTION_CREATE_CATEGORY_RECURSIVE;
      case Constants.FORUM_PERMISSION_ACTION_CREATE_SUBFORUM:
        return Constants.FORUM_PERMISSION_ACTION_CREATE_SUBFORUM_RECURSIVE;
      case Constants.FORUM_PERMISSION_ACTION_CREATE_TOPIC:
        return Constants.FORUM_PERMISSION_ACTION_CREATE_TOPIC_RECURSIVE;
      case Constants.FORUM_PERMISSION_ACTION_CREATE_POST:
        return Constants.FORUM_PERMISSION_ACTION_CREATE_POST_RECURSIVE;
    }
    return -1;
  }

  private Category fixCategory(Content c) {
    Category answer = (Category) c;
    answer.setForum((Forum) c.getParent());
    return answer;
  }

  private Subforum fixSubforum(Content c) {
    Subforum answer = (Subforum) c;
    Content parent = answer.getParent();
    if (parent instanceof Forum) {
      answer.setForum((Forum) parent);
    }
    if (parent instanceof Category) {
      answer.setCategory((Category) parent);
    }
    return answer;
  }

  private Topic fixTopic(Content c) {
    Topic answer = (Topic) c;
    Content parent = answer.getParent();
    if (parent instanceof Forum) {
      answer.setForum((Forum) parent);
    }
    if (parent instanceof Category) {
      answer.setCategory((Category) parent);
    }
    if (parent instanceof Subforum) {
      answer.setSubforum((Subforum) parent);
    }
    return answer;
  }

  private Post fixPost(Content c) {
    Post answer = (Post) c;
    Content parent = answer.getParent();
    if (parent instanceof Forum) {
      answer.setForum((Forum) parent);
    }
    if (parent instanceof Category) {
      answer.setCategory((Category) parent);
    }
    if (parent instanceof Subforum) {
      answer.setSubforum((Subforum) parent);
    }
    if (parent instanceof Topic) {
      answer.setTopic((Topic) parent);
    }
    return answer;
  }

  private List<Content> getContentByParent(String appKey, String parentUid) throws DaliCoreException {
    ContentSearchParams params = new ContentSearchParams();
    Content parent = contentBean.getByUid(appKey, parentUid);
    params.setParent(parent);
    List<Content> search = contentBean.search(appKey, params);
    return search;
  }
}
